/**
 * Java
 * Copyright 2013 IS2T. All rights reserved
 * For demonstration purpose only
 * IS2T PROPRIETARY / Use is subject to license terms.
 */
package com.is2t.examples;

/**
 * This class implements a mutual exclusion monitor that can be used to
 * synchronise a Java thread with one or more C tasks.
 */
public class NativeMonitor {
    private Thread owner;
    private final int id;

    public NativeMonitor() {
        id = newNativeMonitor();
        if (id < 0) {
            throw new IllegalMonitorStateException("Unable to create Native Monitor");
        }
    }

    /**
     * Enters the monitor. Blocks until the monitor can be entered. Once the
     * monitor has been entered no C task or Java thread can enter it until it
     * is released. May be called more than once by the owning thread - a count
     * is kept of the number of calls.
     * 
     * @see exit
     * @throws IllegalMonitorStateException
     *             if the monitor is already taken by another thread.
     */
    public void enter() throws IllegalMonitorStateException {
        synchronized (this) {
            if (owner != null && owner != Thread.currentThread()) {
                throw new IllegalMonitorStateException("Monitor " + id + " is already taken by another thread");
            } else {
                owner = Thread.currentThread();
            }
        }
        while (!tryEnter(id));
    }

    /**
     * Exits the monitor. If the number of times this function has been called
     * equals the number of times that {@link enter} has been called then the
     * monitor becomes available, and if a C task is waiting it is allowed to
     * enter the monitor.
     * 
     * @see enter
     * @throws IllegalMonitorStateException
     *             if the monitor is held by a different thread.
     */
    public void exit() throws IllegalMonitorStateException {
        synchronized (this) {
            if (owner != Thread.currentThread()) {
                throw new IllegalMonitorStateException("Monitor " + id + " is held by a different thread");
            }
            exit(id);
            owner = null;
        }
    }

    /**
     * Provides access to the internal id of the monitor. This id can be passed
     * to a C task so that the C task can access the same monitor.
     * 
     * @return the id of the monitor
     */
    public int getId() {
        return id;
    }

    native private static int newNativeMonitor();
    native private static boolean tryEnter(int id);
    native private static void exit(int id);
}
